"""Regex pattern for parsing the output of an IAR build"""


COMPILER_PATTERN = r"IAR ANSI C/C++ Compiler (V8.50.1.245/W32) for ARM"

# regex pattern for IAR size default output
# Example:
# https://regex101.com/r/5g5Yon/1
# componentA.out
# /opt/iarsystems/bxarm-9.30.1/arm/bin/ilinkarm /src/arm/Debug/Obj/componentA.o \
#   --no_out_extension -o /src/arm/Debug/Exe/componentA.out \
#   --redirect _Printf=_PrintfFullNoMb --redirect _Scanf=_ScanfFullNoMb \
#   --map /src/arm/Debug/List/componentA.map \
#   --config /opt/iarsystems/bxarm-9.30.1/arm/config/generic_cortex.icf \
#   --semihosting /src/arm/Debug/Exe/library.a --entry __iar_program_start \
#   --vfe --text_out locale --cpu=Cortex-M3 --fpu=None

#    IAR ELF Linker V9.30.1.335/LNX for ARM BX
#    Copyright 2007-2022 IAR Systems AB.

#   7'636 bytes of readonly  code memory
#     136 bytes of readonly  data memory
#   1'036 bytes of readwrite data memory

SIZE_PATTERN = (
    r"-o\s.*?(?P<build_output_filename>[\w.]+)\s.*[\n\s]*IAR ELF(.|\n)*?"
    r"(?P<flash_code_size>[\d']+).+readonly\s+code.+\n\s+"
    r"(?P<flash_data_size>[\d']+).+readonly\s+data.+\n\s+"
    r"(?P<ram_size>[\d']+).+readwrite\s+data.+"
)

WARNING_PATTERN = (
    r"(?:\"?(.*?)\"?[\(,](\d+)\)?\s+(?::\s)?)"
    r"(Error|Remark|Warning|Fatal[E|e]rror)\[(.*)\]: (.*)$"
)



"""
`gl_parser_test`
=======================================================================
Unit tests for the parser to pull job contexts from .gitlab-ci.yml files
* Author(s): Bailey Steinfadt
"""
from operator import contains
import pytest

from embedops_cli.yaml_tools import gl_parser, yaml_utilities
from embedops_cli.eo_types import (
    BadYamlFileException,
)
from tests import GLYML_FILENAME

GLYML_MISSING_IMAGE_VALUE_FILENAME = "tests/test-missing-image-value-.gitlab-ci.yml"
GLYML_MISSING_SCRIPT_VALUE_FILENAME = "tests/test-missing-script-value-.gitlab-ci.yml"


def test_get_job_list_with_missing_image_value_exception():
    """test that exception is raised for job with missing image value"""
    with pytest.raises(BadYamlFileException):
        assert gl_parser.get_job_list(GLYML_MISSING_IMAGE_VALUE_FILENAME)


def test_get_job_list_with_missing_script_value_exception():
    """test that exception is raised for job with missing script value"""
    with pytest.raises(BadYamlFileException):
        assert gl_parser.get_job_list(GLYML_MISSING_SCRIPT_VALUE_FILENAME)


def test_get_job_name_list():
    """Test retrieving the list of job names from the YAML"""
    job_name_list = gl_parser.get_job_name_list(GLYML_FILENAME)
    assert len(job_name_list) >= 3
    assert (
        len([job_name for job_name in job_name_list if job_name == "StepScript"]) == 1
    )
    assert (
        len([job_name for job_name in job_name_list if job_name == "StepNoImage"]) == 1
    )
    assert (
        len([job_name for job_name in job_name_list if job_name == "StepNoScript"]) == 1
    )


def test_get_job_list():
    """Test retrieving the list of complete local run context from the YAML"""
    job_list = gl_parser.get_job_list(GLYML_FILENAME)
    assert len(job_list) >= 3
    filtered_job_list = [job for job in job_list if job.job_name == "StepScript"]
    assert len(filtered_job_list) == 1
    assert len(filtered_job_list[0].docker_tag) > 0
    assert len(filtered_job_list[0].script) > 0


def test_get_job_list_default_image():
    """Test retrieving the list of jobs from the YAML,
    check that default image job is as expected"""
    job_list = gl_parser.get_job_list(GLYML_FILENAME)
    job = [job for job in job_list if job.job_name == "StepNoImage"][0]
    assert job.job_name == "StepNoImage"
    assert len(job.docker_tag) > 0
    assert len(job.script) > 0


def test_get_job_list_no_script():
    """Test retrieving the list of jobs from the YAML, check that no script job is as expected"""
    job_list = gl_parser.get_job_list(GLYML_FILENAME)
    job = [job for job in job_list if job.job_name == "StepNoScript"][0]
    assert job.job_name == "StepNoScript"
    assert len(job.docker_tag) > 0
    assert len(job.script) == 0


def test_get_job_context():
    """Test getting the job context for an indicated job"""
    requested_name = "StepScript"
    job = yaml_utilities.get_job_context_for_name(
        gl_parser, GLYML_FILENAME, requested_name
    )
    assert job.job_name == requested_name


def test_get_job_variables():
    """Test getting variables defined in the job"""
    job_list = gl_parser.get_job_list(GLYML_FILENAME)
    job = [job for job in job_list if job.job_name == "StepWithVariable"][0]
    assert "VARYING_THING" in job.variables


def test_get_pipeline_variables():
    """Test getting variables defined outside of job definitions"""
    job_list = gl_parser.get_job_list(GLYML_FILENAME)
    job = [job for job in job_list if job.job_name == "StepWithVariable"][0]
    assert len(job.variables) >= 2
    assert "PROJ_DIR" in job.variables


def test_step_with_literal_multiline_block():
    """Test retrieving a literal multiline block"""
    job = yaml_utilities.get_job_context_for_name(
        gl_parser, GLYML_FILENAME, "StepWithLiteralMultilineBlock"
    )
    assert len(job.script) == 2
    assert job.script[0] == "FILE=.clang-format"
    assert (
        job.script[1]
        == 'if [ -f "$FILE" ]; then echo "$FILE exists. Use repository $FILE."; else echo "$FILE does not exist. Use container $FILE."; cp /tools/.clang-format .clang-format; fi'
    )


def test_step_with_folded_multiline_block():
    """Test retrieving a folded multiline block"""
    job = yaml_utilities.get_job_context_for_name(
        gl_parser, GLYML_FILENAME, "StepWithFoldedMultilineBlock"
    )
    assert len(job.script) == 2
    assert job.script[0] == "FILE=.clang-format"
    assert (
        job.script[1]
        == 'if [ -f "$FILE" ]; then echo "$FILE exists. Use repository $FILE."; else echo "$FILE does not exist. Use container $FILE."; cp /tools/.clang-format .clang-format; fi'
    )


def test_global_var_is_dereferenced():
    """
    Test that when a local variable uses a global one,
    the global var is derefrenced correctly.
    Global:
        PROJ_DIR: "firmware"
    Locals:
        SPECIFIC_DIR: "$PROJ_DIR/spec_dir"
        OTHER_DIR: "${PROJ_DIR}/other_dir"
    """
    job = yaml_utilities.get_job_context_for_name(
        gl_parser, GLYML_FILENAME, "StepWithVariable"
    )
    job_vars = job.variables

    # check the dumb stuff first, everything THERE?
    assert "PROJ_DIR" in job_vars
    assert "SPECIFIC_DIR" in job_vars
    assert "OTHER_DIR" in job_vars

    # Check that the values were dereferenced
    assert job_vars["SPECIFIC_DIR"] == "firmware/spec_dir"
    assert job_vars["OTHER_DIR"] == "firmware/other_dir"



"""
`embedops_cli.utilities`
=======================================================================
miscellaneous utility functions for parsing and handling build logs
* Author(s): Bryan Siepert
"""
from os import getenv
from re import compile as re_compile
import platform
import logging
from sys import exit as sys_exit
import requests

_logger = logging.getLogger(__name__)

compiler_image_regex = re_compile(
    r"^registry\.embedops\.com\/dojofive\/build-images\/"
    r"(?P<compiler>[^\/\:]+)(?:\:(?P<image_version>[^\/]+))?\/?$"
)


def get_compiler():
    """Return the name of the compiler used to generate the build log from either
    `EMBEDOPS_COMPILER`. If that environment variable is not set, we'll try to parse the compiler
    name from the `<CI_IMAGE_NAME_PLACEHOLDER>`"""
    compiler = getenv("EMBEDOPS_COMPILER", default=None)
    if compiler is None:
        _logger.warning("EMBEDOPS_COMPILER was not set. Checking for CI_REGISTRY_IMAGE")
        # fetch compiler from container name, else None
        image_registry_url = getenv("CI_REGISTRY_IMAGE", default=None)
        if image_registry_url is None:
            _logger.error("CI_REGISTRY_IMAGE could not be found")
            sys_exit(1)
        image_match = compiler_image_regex.match(image_registry_url)
        if image_match is None:
            _logger.error(
                f"The Docker image URL provided in {compiler} is not a valid image registry URL"
            )
            sys_exit(1)

        compiler_image = image_match["compiler"].upper()
        image_version = image_match["image_version"].upper()
        _logger.info(
            f"Found compiler image: {compiler_image} image version: {image_version}"
        )
        if "TI" in compiler_image:
            compiler = "TI"
        elif "GCC" in compiler_image:
            compiler = "GCC"
        elif "IAR" in compiler_image:
            compiler = "IAR"
        else:
            _logger.error(f"Compiler {compiler_image} not recognised")
            sys_exit(1)
    return compiler


def post_dict(
    endpoint_uri, data_dict=None, file_dict=None, json_dict=None, headers=None
):
    """POSTs the given object to the given URL as JSON"""

    req = requests.Request(
        "POST",
        endpoint_uri,
        headers=headers,
        files=file_dict,
        json=json_dict,
        data=data_dict,
    )
    prepared = req.prepare()
    session = requests.Session()

    return session.send(prepared)


def quote_str_for_platform(str_to_quote):
    """Properly quote a string of command line arguments for the current platform"""
    if platform.system() == "Windows":
        # Proper quoting for cmd.exe
        #   The entire argument inside double quotes to be taken literally, except
        #   for double quotes inside the string literal, which must be escaped.
        # To accomplish this, we replace all instances of '"' with '\"'.
        #   Note: '\' must be used to escape itself in the replace pattern.
        str_to_quote = str_to_quote.replace('"', '\\"')
        return f'"{str_to_quote}"'

    # Proper quoting for Bash/zsh/etc.
    #   The entire argument inside single quotes to be taken literally.
    return f"'{str_to_quote}'"



"""
`embedops_cli`
=======================================================================
CLI interface for EmbedOps tools
* Author(s): Bailey Steinfadt
"""
import sys
import logging
import click

# import embedops_cli.yaml_tools.yaml_utilities as yaml_utilities
from embedops_cli.yaml_tools import yaml_utilities
from . import docker_run, version, embedops_authorization, telemetry
from .eo_types import (
    BadYamlFileException,
    DockerNotRunningException,
    EmbedOpsException,
    LoginFailureException,
    NoYamlFileException,
    UnsupportedYamlTypeException,
    MultipleYamlFilesException,
    NoDockerContainerException,
    InvalidDockerContainerException,
    UnauthorizedUserException,
    DockerRegistryException,
    UnknownDockerException,
)

_logger = logging.getLogger(__name__)

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help", "--halp"])


@click.group(
    invoke_without_command=True,
    context_settings=CONTEXT_SETTINGS,
)
@click.version_option(version=version.__version__)
@click.option("--debug", "-d", is_flag=True, help="Enable debug logging")
@click.pass_context
def embedops_cli(ctx, debug):
    """EmbedOps Base Command"""

    if debug:
        logging.basicConfig(level=logging.DEBUG)
        _logger.debug("Debug logging enabled")
    else:
        logging.basicConfig(level=logging.INFO)

    if ctx.invoked_subcommand is None:
        click.secho("-" * 80, fg="magenta")
        click.secho(
            "\n╭━━━╮╱╱╭╮╱╱╱╱╱╱╭┳━━━╮\n"
            "┃╭━━╯╱╱┃┃╱╱╱╱╱╱┃┃╭━╮┃\n"
            "┃╰━━┳╮╭┫╰━┳━━┳━╯┃┃╱┃┣━━┳━━╮\n"
            "┃╭━━┫╰╯┃╭╮┃┃━┫╭╮┃┃╱┃┃╭╮┃━━┫\n"
            "┃╰━━┫┃┃┃╰╯┃┃━┫╰╯┃╰━╯┃╰╯┣━━┃\n"
            "╰━━━┻┻┻┻━━┻━━┻━━┻━━━┫╭━┻━━╯\n"
            "╱╱╱╱╱╱╱╱╱╱╱╱╱╱╱╱╱╱╱╱┃┃\n"
            "╱╱╱╱╱╱╱╱╱╱╱╱╱╱╱╱╱╱╱╱╰╯\n",
            fg="magenta",
        )
        click.secho(
            f"\nWelcome to EmbedOps CLI",
            err=False,
            fg="magenta",
        )
        click.secho("Version: " + version.__version__ + "\n")
        click.secho(
            "EmbedOps consists of tools, templates, and services that focus on the foundation\n"
            "of any software development environment. It provides web-based and command-line\n"
            "tools that make setting up and maintaining your builds smooth and simple.\n"
            "EmbedOps tools also integrate directly with your automated CI pipelines, allowing\n"
            "any developer to run any step in the production CI pipeline in their local dev\n"
            "environment exactly as it would be run on the CI server.\n"
        )
        click.secho(
            "Example:\n"
            '"embedops-cli jobs show" provides a listing of the jobs that EmbedOps can run\n'
        )
        click.secho(
            "For a listing of all options, use embedops-cli --help, or embedops-cli -h\n"
        )
        click.secho("-" * 80, fg="magenta")
        click.secho("\n")


def _say_token_is_good(token_name: str):
    click.secho(f"\n{token_name} Token is ", nl=False)
    click.secho("GOOD", err=False, fg="bright_green")
    click.secho("You are logged into EmbedOps!\n", err=False, fg="white")


@embedops_cli.command(hidden=True)
@click.option("--test", "-t", help="Test your login status", is_flag=True)
def login(test):
    """Log into the EmbedOps platform.
    You will be prompted to enter your EmbedOps credentials if you are not logged in."""
    token = embedops_authorization.get_auth_token()

    if token and embedops_authorization.check_token():
        _say_token_is_good("EmbedOps")
    elif test:
        click.secho("\nToken not found", err=False, fg="bright_red")
        click.secho(
            "\nuse `embedops-cli login` to log in and retrieve a token",
            err=False,
            fg="bright_red",
        )
        sys.exit(1)
    else:
        try:  # request a token if we don't already have one
            access_token = embedops_authorization.request_authorization()
            if access_token is None:
                raise LoginFailureException()
            embedops_authorization.set_auth_token(access_token)

        except (LoginFailureException, UnauthorizedUserException) as exc:
            _echo_error_and_fix(exc)
            sys.exit(1)

    try:
        embedops_authorization.fetch_registry_token()
        embedops_authorization.login_to_registry()
    except (
        UnauthorizedUserException,
        LoginFailureException,
        DockerRegistryException,
        UnknownDockerException,
    ) as exc:
        _echo_error_and_fix(
            exc
        )  # TODO: add context to these so we know WHICH login failed
    except DockerNotRunningException as exc:
        _echo_error_and_fix(exc)
    else:
        _say_token_is_good("Docker")

    telemetry.login_event()


@embedops_cli.group()
@click.option(
    "--filename",
    help="path to the CI YAML or YML file",
    required=False,
    expose_value=True,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
)
@click.pass_context
def jobs(ctx: click.Context, filename):
    """Run or view CI jobs defined in
    YAML or YML files locally"""

    if filename is None:
        try:
            filename = yaml_utilities.get_yaml_in_directory()
        except (NoYamlFileException, MultipleYamlFilesException) as exc:

            _echo_error_and_fix(exc)

            click.secho(ctx.get_usage(), err=True, fg="white")
            sys.exit(2)
    else:
        if not (filename.lower().endswith(".yaml") or filename.endswith(".yml")):
            click.secho("-" * 80, fg="bright_red")
            click.secho("File must be a .yaml or .yml file.", err=True, fg="bright_red")
            click.secho(ctx.get_usage(), err=True, fg="white")
            click.secho("-" * 80, fg="bright_red")
            sys.exit(1)
    ctx.obj = filename


@jobs.command()
@click.pass_context
@click.argument("job_name")
def run(ctx: click.Context, job_name):
    """Run a job defined in a CI YAML file.
    JOB_NAME is the name of the job or step in your CI YAML file"""

    telemetry.command_event("jobs_run", {"job_name": job_name})

    # check if token has expired if so try to get a new token
    if embedops_authorization.is_registery_token_valid() is False:
        embedops_authorization.login_to_registry()

    filename = ctx.obj
    _logger.debug(f"jobs run called with file '{filename}' and job '{job_name}")

    try:
        job_list = yaml_utilities.get_job_list(filename)
    except (
        UnsupportedYamlTypeException,
        BadYamlFileException,
    ) as exc:
        _echo_error_and_fix(exc)
        ctx.exit(2)

    # match the given job name against the job collection
    for job in job_list:
        if job_name == job.job_name:
            click.secho("-" * 80, fg="magenta")
            click.secho(
                f"Running job '{job_name}'",
                err=False,
                fg="magenta",
            )
            click.secho(
                f"from file '{filename}'\n" f"in image '{job.docker_tag}'",
                err=False,
                fg="white",
            )
            click.secho("-" * 80, fg="magenta")
            click.secho("\n")
            try:
                run_exitcode = docker_run.docker_run(job)
            except (
                EmbedOpsException,
                NoDockerContainerException,
                InvalidDockerContainerException,
            ) as exc:
                _echo_error_and_fix(exc)
                ctx.exit(1)
            if run_exitcode == 0:
                click.secho("\nJob ran successfully\n", err=False, fg="magenta")
            else:
                click.secho("\nJob ran with errors\n", err=True, fg="red")
            break
    # They tried to run a job that doesn't exist, show them the jobs they can run.
    else:
        try:
            job_name_list = "\n".join(_get_job_name_list(filename))
            click.secho(
                f'\nJob "{job_name}" is not available in this CLI configuration.\n',
                err=False,
                fg="yellow",
            )
            click.secho(
                f"EmbedOps CLI Jobs Available:",
                err=False,
                fg="magenta",
            )
            click.secho(
                f"{job_name_list}\n",
                err=False,
                fg="white",
            )
        except (UnsupportedYamlTypeException, BadYamlFileException) as exc:
            _echo_error_and_fix(exc)
            ctx.exit(2)


@jobs.command()
@click.option(
    "-v",
    "--verbose",
    help="Show details for the available jobs in YAML file",
    required=False,
    expose_value=True,
    is_flag=True,  # Inform Click that this is a boolean flag
)
@click.pass_context
def show(ctx: click.Context, verbose):
    """Show available jobs in YAML file"""

    telemetry.command_event("jobs_show", {"verbose": verbose})

    filename = ctx.obj
    _logger.debug(f"jobs show called with file {filename}")

    if not verbose:
        try:
            job_name_list = "\n".join(_get_job_name_list(filename))
        except (UnsupportedYamlTypeException, BadYamlFileException) as exc:
            _echo_error_and_fix(exc)
            ctx.exit(2)

        click.secho(f"\nEmbedOps CLI Jobs Available:", err=False, fg="magenta")
        click.secho(f"{job_name_list}\n", err=False, fg="white")
    else:
        try:
            job_list = yaml_utilities.get_job_list(filename)
        except (
            UnsupportedYamlTypeException,
            BadYamlFileException,
        ) as exc:
            _echo_error_and_fix(exc)
            ctx.exit(2)

        click.secho("\nEmbedOps CLI Jobs Details:\n", err=False, fg="magenta")

        # match the given job name against the job collection
        for job in job_list:
            click.secho(
                f"{job.job_name}\n  Image: {job.docker_tag}\n  Variables:",
                err=False,
                fg="white",
            )
            for variable, value in job.variables.items():
                click.secho(f"    {variable}: {value}", err=False, fg="white")
            click.secho(f"  Script:", err=False, fg="white")
            for line in job.script:
                click.secho(f"    - {line}", err=False, fg="white")
            click.secho("", err=False, fg="white")


@jobs.command()
@click.pass_context
@click.argument("job_name")
def describe(ctx: click.Context, job_name):
    """Shows details for a single job"""

    telemetry.command_event("jobs_describe", {"job_name": job_name})

    filename = ctx.obj
    _logger.debug(f"jobs show called with file {filename}")

    try:
        job = yaml_utilities.get_job_context_for_name(
            yaml_utilities.get_correct_parser_type(filename), filename, job_name
        )
    except (
        UnsupportedYamlTypeException,
        BadYamlFileException,
    ) as exc:
        _echo_error_and_fix(exc)
        ctx.exit(2)

    # match the given job name against the job collection
    if job is not None:
        click.secho(
            f"\n{job.job_name}",
            err=False,
            fg="magenta",
        )
        click.secho(
            f"Image: {job.docker_tag}\n  Variables:",
            err=False,
            fg="white",
        )
        for variable, value in job.variables.items():
            click.secho(f"    {variable}: {value}", err=False, fg="white")
        click.secho(f"  Script:", err=False, fg="white")
        for line in job.script:
            click.secho(f"    - {line}", err=False, fg="white")
        click.secho("", err=False, fg="white")
    # They tried to show details of a job that doesn't exist, show them the jobs they can run.
    else:
        try:
            job_name_list = "\n    ".join(_get_job_name_list(filename))

            click.secho(
                (
                    f"  Job {job_name} is not available in this CI config.\n"
                    "  Jobs available:\n"
                    f"    {job_name_list}"
                ),
                err=False,
                fg="white",
            )
        except (UnsupportedYamlTypeException, BadYamlFileException) as exc:
            _echo_error_and_fix(exc)
            ctx.exit(2)


def _get_job_name_list(_filename: str) -> list:
    try:
        parser = yaml_utilities.get_correct_parser_type(_filename)
    except UnsupportedYamlTypeException as exc:
        raise UnsupportedYamlTypeException() from exc

    try:
        return parser.get_job_name_list(_filename)
    except BadYamlFileException as exc:
        raise BadYamlFileException() from exc


def _echo_error_and_fix(exc):
    """Print the error message and fix messages on exceptions"""
    click.secho("-" * 80, fg="bright_red")
    click.secho(exc.message, err=True, fg="bright_red")
    click.secho(exc.fix_message, err=True, fg="white")
    click.secho("-" * 80 + "\n", fg="bright_red")


if __name__ == "__main__":
    embedops_cli(prog_name="embedops-cli")  # pylint:disable=unexpected-keyword-arg



"""
Group of tests for the embedops_cli and eo_tools
"""

GLYML_FILENAME = "tests/test-.gitlab-ci.yml"
BBYML_FILENAME = "tests/test-bitbucket-pipelines.yml"
GHYML_FILENAME = "tests/test-.github-ci.yml"
